from flask import Blueprint, Markup, Response, render_template, request, url_for
from . import menu, param
from .ros import *

bp = Blueprint('node', __name__, url_prefix='/node')

@bp.route('/')
def get_node_overview():
    # Update ros api
    ros.update()

    # Get menu items
    menu_items = menu.get_items()

    # Get list of nodes
    content = ''
    items = menu_items['nav_node_items']

    # Return rendered template
    return render_template('base_with_list.html', title=f'List of Nodes',
                            active_menu_item='node',
                            content=content,
                            items=items,
                            **menu_items)    

@bp.route('/<path:name>')
def get_node_info(name):
    # Check if a svg representation should be returned
    generate_svg = request.args.get('get', '', type=str) == 'svg'

    if not name.startswith('/'):
        name = '/' + name

    # Update ros api
    ros.update()

    # Get node
    node = ros.get_node(name)

    if generate_svg is True:
        # Generate graph
        svg = node.svg()
        return Response(svg, mimetype='image/svg+xml')
    else:
        # Assemble content
        content = ''
        param_content = param.get_param(name)
        if param_content is not None:
            content += Markup('<h2>Parameter</h2>')
            content += param.get_param(name)

        # Link to svg
        url = url_for('node.get_node_info', name=name)
        img_data = url + '?get=svg'

        # # Embed image directly as svg in object
        # For some reason the browsers seem to block the links with this version
        # graph = get_graph(data, name)
        # img_stream = BytesIO()
        # graph.draw(path=img_stream, format='svg', prog='dot')
        # svg = img_stream.getvalue().decode('utf-8')
        # svg = svg.replace('xlink:', '')
        # message_bytes = svg.encode('ascii')
        # base64_bytes = base64.b64encode(message_bytes)
        # base64_message = base64_bytes.decode('ascii')
        # img_data = 'data:image/svg+xml;base64,' + base64_message

        # Return rendered template
        return render_template('base_with_svg.html', title=f'Node {name}',
                               active_menu_item='node',
                               content=content,
                               **menu.get_items(active_item=url),
                               img_data=img_data)
                            


from flask import Flask, Response, render_template
from . import node, service, topic, param

def create_app(config=None):
    import os
    import rospy
    import signal
    import sys
    import yaml
    from . import menu
    from .ros import ros

    # Signal handler (don't know why this is needed)
    def signal_handler(sig, frame):
        print('You pressed Ctrl+C!')
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    # Read config file
    if config is None:
        config_file = os.path.join(os.path.dirname(__file__), 'config.yaml')
        with open(config_file, 'r') as stream:
            try:
                config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)

    # Instantiate ROS api
    rospy.init_node('ros_web_gui')
    if 'blacklisted_nodes' not in config:
        config['blacklisted_nodes'] = ['ros_web_gui']
    else:
        config['blacklisted_nodes'].append('ros_web_gui')
        
    # Configure ROS api
    ros.config(**config)

    # Start ROS api
    ros.update()

    # Assemble app
    app = Flask(__name__)
    app.register_blueprint(node.bp)
    app.register_blueprint(service.bp)
    app.register_blueprint(topic.bp)
    app.register_blueprint(param.bp)

    @app.route('/graph.svg')
    def get_graph_svg():
        # Update ros api
        ros.update()

        # Get svg
        svg = ros.svg()
        return Response(svg, mimetype='image/svg+xml')

    @app.route('/')
    @app.route('/index.html')
    def get_main():
        # Update ros api
        ros.update()

        # Empty content at the moment
        content = ''

        # Link to svg
        img_data = 'graph.svg'

        # Render and return template
        return render_template('base_with_svg.html', title='Overview',
                               active_menu_item='home',
                               content=content, **menu.get_items(),
                               img_data=img_data)

    return app


from .ros import ros
from flask import url_for

def get_items(active_item=None):
    # Navigation
    nav_node_items =    [{'href': url_for('node.get_node_info', name=node), 'caption': node} for node in sorted(ros.nodes)]
    nav_topic_items =   [{'href': url_for('topic.get_topic_info', name=topic), 'caption': topic} for topic in sorted(ros.topics)]
    nav_service_items = [{'href': url_for('service.get_service_info', name=service), 'caption': service} for service in sorted(ros.services)]
    nav_param_items =   [{'href': url_for('param.get_param_info', name=param), 'caption': param} for param in sorted(ros.params)]

    if active_item is not None:
        nav_node_items = [{**item, 'active': True if item['href'] == active_item else False} for item in nav_node_items]
        nav_topic_items = [{**item, 'active': True if item['href'] == active_item else False} for item in nav_topic_items]
        nav_service_items = [{**item, 'active': True if item['href'] == active_item else False} for item in nav_service_items]
        nav_param_items = [{**item, 'active': True if item['href'] == active_item else False} for item in nav_param_items]

    return {
        'nav_node_items': nav_node_items,
        'nav_topic_items': nav_topic_items,
        'nav_service_items': nav_service_items,
        'nav_param_items': nav_param_items
    }



from flask import Blueprint, Markup, Response, jsonify, render_template, request, url_for
from . import menu
from .ros import ros

bp = Blueprint('topic', __name__, url_prefix='/topic')

@bp.route('/')
def get_topic_overview():
    # Update ros api
    ros.update()

    # Get menu items
    menu_items = menu.get_items()

    # Iterate over nodes
    content = ''
    items = menu_items['nav_topic_items']

    # Return rendered template
    return render_template('base_with_list.html', title=f'List of Topics',
                            active_menu_item='topic',
                            content=content,
                            items=items,
                            **menu_items)   

@bp.route('/<path:name>', methods=['GET', 'POST'])
def get_topic_info(name):
    # Check if a json or svg representation should be returned
    mode = request.args.get('get', '', type=str)

    # Update ros api
    ros.update()

    # Add slash to name
    if name[0] != '/':
        name_without_slash = name
        name = '/' + name
    else:
        name_without_slash = name[1:]

    # Get topic
    topic = ros.get_topic(name)

    # Handle post request
    if request.method == 'POST':
        msg_pub = request.form['msg_pub']
        response = {
            'success': True,
            'error_msg': None
        }
        try:
            topic.publish(msg_pub)
        except Exception as e:
            response['success'] = False
            response['error_msg'] = str(e)
        return jsonify(response)

    # Return svg
    if mode == 'svg':
        svg = topic.svg()
        return Response(svg, mimetype='image/svg+xml')

    # Get message
    msg = topic.msg
    if msg is not None:
        msg_data = [
            {'id': 'num_messages', 'description': 'Messages received', 'value': msg['num_messages']},
            {'id': 'messages_per_second', 'description': 'Messages per second', 'value': f"{msg['messages_per_second']:.1f}"},
            {'id': 'avg_size', 'description': 'Average message size', 'value': f"{msg['avg_size']:.0f} Bytes"},
            {'id': 'bandwidth', 'description': 'Bandwidth', 'value': f"{msg['avg_size']*msg['messages_per_second']:.0f} Bytes/s"},
            {'id': 'last_message', 'description': 'Last message received', 'value': msg["last_message"].strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]},
            {'id': 'msg', 'description': 'Message', 'value': str(msg['msg'])}
        ]           
    else:
        msg_data = [
            {'id': 'num_messages', 'description': 'Messages received', 'value': 0},
            {'id': 'messages_per_second', 'description': 'Messages per second', 'value': '-'},
            {'id': 'avg_size', 'description': 'Average message size', 'value': '-'},
            {'id': 'bandwidth', 'description': 'Bandwidth', 'value': '-'},
            {'id': 'last_message', 'description': 'Last message received', 'value': '-'},
            {'id': 'msg', 'description': 'Message', 'value': 'None'}
        ]

    if mode == 'json':
        print(f'Returning JSON for topic {topic.name}')
        return jsonify(msg_data)
    else:
        content = ''  # get_topic_info_description(topic)

        # Format topic info
        content = Markup(content.replace('\n', '<br/>'))

        # Image url
        url = url_for('topic.get_topic_info', name=name_without_slash)
        img_data = url + '?get=svg'

        # Return rendered template
        return render_template('topic.html', title=f'Topic {name}',
                               active_menu_item='topic',
                               url=url,
                               msg_items=msg_data,
                               msg=None if msg is None else msg['msg'],
                               msg_template=topic.msg_template(),
                               content=content,
                               **menu.get_items(active_item=url),
                               img_data=img_data)
                            


import roslib
import rospy
import sys
import yaml

def get_object_size(obj, seen=None):
    """Recursively finds size of objects
    Copied from https://goshippo.com/blog/measure-real-size-any-python-object/
    """
    size = sys.getsizeof(obj)
    
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_object_size(v, seen) for v in obj.values()])
        size += sum([get_object_size(k, seen) for k in obj.keys()])

    if hasattr(obj, '__dict__'):
        size += get_object_size(obj.__dict__, seen)

    if hasattr(obj, '__slots__'):
        size += get_object_size(obj.__slots__, seen)

    if hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_object_size(i, seen) for i in obj])
    return size

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

def _dict_to_msg(data):
    if isinstance(data, dict):    
        # ToDo: Not really a nice special case for timestamps
        if 'secs' in data and 'nsecs' in data:
            return rospy.get_rostime()
        data = AttrDict(**data)
        for key in data:
            data[key] = _dict_to_msg(data[key])
        return data
    elif isinstance(data, list):
        for idx, ele in enumerate(data):
            data[idx] = _dict_to_msg(ele)
        return data
    else:
        return data

def str_to_msg(data, cls):
    data_dict = yaml.safe_load(data)
    if data_dict is None or isinstance(data_dict, str):
        return cls()
    msg = _dict_to_msg(data_dict)
    obj = cls(**msg)
    return obj

def initialize_msg(msg):
    primitives = {
        "bool": False,
        "int8": 0,
        "uint8": 0,
        "int16": 0,
        "uint16": 0,
        "int32": 0,
        "uint32": 0,
        "int64": 0,
        "uint64": 0,
        "float32": 0.0,
        "float64": 0.0,
        "string": "",
        #"time": rospy.Time.now(),
        #"duration": rospy.Duration()
    }
    
    for idx, name in enumerate(msg.__slots__):
        msg_type = msg._slot_types[idx]
        if msg_type.endswith('[]'):
            msg_type = msg_type[:-2]
            if msg_type in primitives:
                setattr(msg, name, [primitives[msg_type]])
            else:
                msg_class = roslib.message.get_message_class(msg_type[:-2])
                setattr(msg, name, [initialize_msg(msg_class())])
        elif msg_type in primitives:
            setattr(msg, name, primitives[msg_type])
        else:
            msg_class = roslib.message.get_message_class(msg_type)
            setattr(msg, name, initialize_msg(msg_class()))
    return msg


from flask import Blueprint, Markup, Response, jsonify, render_template, request, url_for
from . import menu
from .ros import ros
from .util import str_to_msg, initialize_msg

bp = Blueprint('service', __name__, url_prefix='/service')

@bp.route('/')
def get_service_overview():
    # Update ros api
    ros.update()

    # Get menu items
    menu_items = menu.get_items()

    # Iterate over nodes
    content = ''
    items = menu_items['nav_service_items']

    # Return rendered template
    return render_template('base_with_list.html', title=f'List of Services',
                            active_menu_item='service',
                            content=content,
                            items=items,
                            **menu_items)   

@bp.route('/<path:name>', methods=['GET', 'POST'])
def get_service_info(name):
    # Add slash to name
    if name[0] != '/':
        name_without_slash = name
        name = '/' + name
    else:
        name_without_slash = name[1:]        

    # Update ros api
    ros.update()

    # Get service
    service = ros.get_service(name)

    # Handle post request
    if request.method == 'POST':
        print("Got service call request")
        req_msg = request.form['request']
        ret = {
            'success': True,
            'error_msg': None,
            'response': None
        }
        try:
            req = str_to_msg(req_msg, service.request_class)
            res = service.call(req)
            ret['response'] = str(res)
        except Exception as e:
            ret['success'] = False
            ret['error_msg'] = str(e)
        return jsonify(ret)

    # Check if a svg representation should be returned
    generate_svg = request.args.get('get', '', type=str) == 'svg'    

    if generate_svg is True:
        svg = service.svg()
        return Response(svg, mimetype='image/svg+xml')
    else:
        # # Get service info
        # old_stdout = sys.stdout
        # sys.stdout = mystdout = StringIO()
        # rosservice._rosservice_info(name)
        # content = mystdout.getvalue()
        # sys.stdout = old_stdout
        # mystdout.close()
        content = ''

        # Format node info
        content = Markup(content.replace('\n', '<br/>'))

        # Link to svg
        url = url_for('service.get_service_info', name=name_without_slash)
        img_data = url + '?get=svg'

        # Return rendered template
        return render_template('service.html', title=f'Service {name}',
                               active_menu_item='service',
                               url=url,
                               content=content,
                               msg_template=initialize_msg(service.request_class()),
                               **menu.get_items(active_item=url),
                               img_data=img_data)
                               #img_data=img_stream)
                               #img_data=urllib.parse.quote(img_stream.rstrip('\n')))


from flask import Blueprint, Markup, escape, render_template, url_for
from . import menu
from .ros import ros
import rosgraph
import rosparam

bp = Blueprint('param', __name__, url_prefix='/param')

def get_text(content, namespace='/'):
    buff = ''
    if type(content) == list:
        for idx, ele in enumerate(content):
            if type(ele) is dict:
                buff += get_text(ele, namespace + f'[{idx}]/')
            else:
                buff += get_text(ele, namespace + f'[{idx}]')
    elif type(content) == dict:
        for key, val in content.items():
            if type(val) is not dict:
                new_namespace = namespace + f'{key}'
            else:
                new_namespace = namespace + f'{key}/'
            buff += get_text(val, new_namespace)
    else:
        buff += f'{namespace}<br />'
        if type(content) == str and len(content) > 5 and content[:5] == '<?xml':
            buff +=  f'<pre style="display: inline-block"><code class="html">{escape(content)}</code></pre><br />\n'
        else:
            buff +=  f'<pre style="display: inline-block"><code class="plaintext">{escape(content)}</code></pre><br />\n'
    return buff

def get_param(name):
    # Generate content
    try:
        content = rosparam.get_param(name)
    except rosparam.RosParamIOException:
        #content = f'Could net get value of {name} from parameter server'
        return None
    except rosgraph.masterapi.MasterError:
        return None

    # Ensure the namespace ends with a slash
    namespace = name if name[-1] == '/' else name + '/'

    # Type handling
    content = get_text(content, namespace=namespace)

    # Format param info
    content = Markup('<div class="w3-container">' + content + '</div>')
    return content

# @bp.route('/')
# def get_param_overview():
#     # Get info about nodes
#     data = ros.get_info()

#     # Get menu items
#     menu_items = menu.get_items(data)

#     # Iterate over nodes
#     content = ''
#     items = menu_items['nav_param_items']

#     # Return rendered template
#     return render_template('base_with_list.html', title=f'List of Parameters',
#                             active_menu_item='param',
#                             content=content,
#                             items=items,
#                             **menu_items)

@bp.route('/')
def get_param_overview():
    # Update ros api
    ros.update()

    # Get menu items
    menu_items = menu.get_items()

    # Iterate over nodes
    content = get_param('/')

    # Return rendered template
    return render_template('base.html', title=f'List of Parameters',
                            active_menu_item='param',
                            content=content,
                            **menu_items)

@bp.route('/<path:name>')
def get_param_info(name):
    if not name.startswith('/'):
        name = '/' + name

    # Update ros api
    ros.update()

    # Generate content
    content = get_param(name)

    # Current url
    url = url_for('param.get_param_info', name=name)

    # Return rendered template
    return render_template('base.html', title=f'Parameter {name}',
                            active_menu_item='param',
                            content=content,
                            **menu.get_items(active_item=url))
                            


import re
import setuptools

with open("ros_web_gui/__init__.py", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ros_web_gui",
    version=version,
    author="Simon Manschitz",
    author_email="simon.manschitz@gmx.de",
    description="A web gui for ROS (robot operating system)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/smanschi/ros_web_gui",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    scripts=['scripts/ros_web_gui'],
    package_data={'ros_web_gui': ['*', 'static/*', 'templates/*']},
    include_package_data=True,
    install_requires=[
        'Flask>=1.1.2',
        'PyYAML>=5.3.1',
        'pydot>=1.4.2',
        'rosdep>=0.20.1'
    ]
)



import genpy
import pydot
import rosgraph
import roslib
import rosnode
import rosparam
import rospy
import rosservice
import rostopic
import socket
from datetime import datetime
from flask import Markup, url_for
from io import BytesIO
from .util import get_object_size, str_to_msg, initialize_msg


def graph_to_svg(graph):
    img_str = graph.create_svg(prog='dot')
    sio = BytesIO()
    sio.write(img_str)
    sio.seek(0)
    svg = sio.getvalue().decode('utf-8')
    #svg = svg.replace('xlink:', '')
    svg = Markup(svg.replace('xlink:', '').replace('w3&#45;', 'w3-').replace('hover&#45;', 'hover-'))
    return svg


class Topic():
    def __init__(self, name):
        super().__init__()
        rospy.loginfo(f"Instantiating topic with name {name}")
        self.__name = name
        self.__msg = None
        self.__msg_info = None
        try:
            topic_type_info = rostopic.get_topic_type(name)
            self.__type = topic_type_info[0]
        except Exception as e:
            rospy.logerr(str(e))
            self.type = None
        try:
            self.__data_class = roslib.message.get_message_class(topic_type_info[0])
            self.__msg_template = initialize_msg(self.__data_class())
            self.__ros_subscriber = rospy.Subscriber(name, self.__data_class, self.__onMessage)
            self.__ros_publisher = rospy.Publisher(name, self.__data_class, queue_size=1)
        except Exception as e:
            rospy.logerr(str(e))
            self.__data_class = None
            self.__ros_subscriber = None
            self.__ros_publisher = None
        self.__subs = dict()
        self.__pubs = dict()
        self.__graph = None
        self.__svg = None
        self.__times = {
            'state': datetime.now(),
            'graph': None,
            'svg': None
        }

    @property
    def name(self):
        return self.__name

    @property
    def type(self):
        return self.__type

    @property
    def msg(self):
        if self.__msg_info is None:
            return None
        return {
            'msg': self.__msg,
            'num_messages': self.__msg_info['num_messages'],
            'messages_per_second': float(self.__msg_info['num_messages']) / (datetime.now() - self.__msg_info['first_message']).total_seconds(),
            'last_message': self.__msg_info['last_message'],
            'avg_size': self.__msg_info['avg_size']
        }

    @property
    def data_class(self):
        return self.__data_class

    @property
    def subscribers(self):
        return self.__subs

    @property
    def publishers(self):
        return self.__pubs

    def clear(self):
        self.__pubs.clear()
        self.__subs.clear()
        self.__times['state'] = datetime.now()

    def unregister(self):
        if self.__ros_subscriber is not None:
            self.__ros_subscriber.unregister()
        if self.__ros_publisher is not None:
            self.__ros_publisher.unregister()

    def addPublisher(self, node):
        self.__pubs[node.name] = node
        self.__times['state'] = datetime.now()

    def addSubscriber(self, node):
        self.__subs[node.name] = node
        self.__times['state'] = datetime.now()

    def publish(self, msg_str):
        try:
            msg = str_to_msg(msg_str, self.__ros_publisher.data_class)
            self.__ros_publisher.publish(msg)
        except Exception as e:
            rospy.loginfo(e)
            raise e

    def graph(self):
        # Check if we can return an existing graph
        if (self.__graph is not None) and (self.__times['graph'] >= self.__times['state']):
            return self.__graph

        # Remember time
        rospy.loginfo(f'Generating graph for topic {self.__name}')
        self.__times['graph'] = datetime.now()

        # Generate new graph
        graph = pydot.Dot(graph_type='digraph', forcelabels=True)
        topic_id = 'topic_' + self.__name
        topic_label = f"{self.__name}\\n{self.__type}"
        graph.add_node(pydot.Node(topic_id, label=topic_label, shape='box'))

        # Add publishers
        for node_name in self.__pubs:
            node_id = 'pub_' + node_name
            node_url = url_for('node.get_node_info', name=node_name)
            graph.add_node(pydot.Node(node_id, label=node_name, shape='oval', URL=node_url, target='_top'))
            graph.add_edge(pydot.Edge(node_id, topic_id))

        # Add subscribers
        for node_name in self.__subs:
            node_id = 'sub_' + node_name
            node_url = url_for('node.get_node_info', name=node_name)
            graph.add_node(pydot.Node(node_id, label=node_name, shape='oval', URL=node_url, target='_top'))
            graph.add_edge(pydot.Node(topic_id, node_id))

        # Store and return graph
        self.__graph = graph
        return graph

    def svg(self):
        # Check if we can return an existing svg
        graph = self.graph()
        if (self.__svg is not None) and (self.__times['svg'] >= self.__times['graph']):
            return self.__svg

        # Remember time
        rospy.loginfo(f'Generating svg for topic {self.__name}')
        self.__times['svg'] = datetime.now()

        # Store and return svg
        self.__svg = graph_to_svg(graph)
        return self.__svg

    def msg_template(self):
        return genpy.message.strify_message(self.__msg_template)

    def __onMessage(self, msg):
        if self.__msg_info is None:
            self.__msg_info = {
                'num_messages': 1,
                'first_message': datetime.now(),
                'last_message': datetime.now(),
                'avg_size': get_object_size(msg)          
            }
        else:
            self.__msg_info['num_messages'] += 1
            self.__msg_info['last_message'] = datetime.now()
            self.__msg_info['avg_size'] = (self.__msg_info['avg_size']*(self.__msg_info['num_messages']-1) + get_object_size(msg))/self.__msg_info['num_messages']

        self.__msg = msg


class Node:
    def __init__(self, name):
        self.__name = name
        self.__subs = dict()
        self.__pubs = dict()
        master = rosgraph.Master(name)
        self.__uri = master.lookupNode(name)
        self.__srvs = dict()
        self.__graph = None
        self.__svg = None
        self.__times = {
            'state': datetime.now(),
            'graph': None,
            'svg': None
        }

    @property
    def name(self):
        return self.__name

    @property
    def subscriptions(self):
        return self.__subs

    @property
    def publications(self):
        return self.__pubs

    @property
    def services(self):
        return self.__srvs

    @property
    def uri(self):
        return self.__uri

    def addSubscription(self, sub):
        self.__subs[sub.name] = sub
        self.__times['state'] = datetime.now()

    def addPublication(self, pub):
        self.__pubs[pub.name] = pub
        self.__times['state'] = datetime.now()

    def addService(self, srv):
        self.__srvs[srv.name] = srv
        self.__times['state'] = datetime.now()

    def graph(self):
        # Check if we can return an existing graph
        if (self.__graph is not None) and (self.__times['graph'] >= self.__times['state']):
            return self.__graph

        # Remember time
        rospy.loginfo(f'Generating graph for node {self.__name}')
        self.__times['graph'] = datetime.now()

        # Generate new graph
        graph = pydot.Dot(graph_type='digraph', forcelabels=True, stylesheet='https://www.w3schools.com/w3css/4/w3.css')

        # Add node
        node_label = f"{self.__name}\\n{self.__uri}"
        graph.add_node(pydot.Node(self.__name, label=node_label, **{'shape': 'oval', 'class': 'w3-orange w3-hover-red'}))

        # Add topics to which the node subscribes
        node_names = set()
        for topic_name, topic in self.__subs.items():
            topic_id = 'topic_sub_' + topic_name
            node_url = url_for('topic.get_topic_info', name=topic_name)
            node_label = f"{topic_name}\\n{topic.type}"
            graph.add_node(pydot.Node(topic_id, label=node_label, shape='box', URL=node_url, target='_top'))
            graph.add_edge(pydot.Edge(topic_id, self.__name))

            for pub_name in topic.publishers:
                subnode_id = 'pub_' + pub_name
                if pub_name not in node_names:
                    subnode_url = url_for('node.get_node_info', name=pub_name)
                    node_label = f"{pub_name}\\n{topic.publishers[pub_name].uri}"
                    graph.add_node(pydot.Node(subnode_id, label=node_label, shape='oval', URL=subnode_url, target='_top'))
                    node_names.add(pub_name)
                graph.add_edge(pydot.Edge(subnode_id, topic_id))

        # Add topics which the node publishes
        node_names.clear()
        for topic_name, topic in self.__pubs.items():
            topic_id = 'topic_pub_' + topic_name
            node_url = url_for('topic.get_topic_info', name=topic_name)
            node_label = f"{topic_name}\\n{topic.type}"
            graph.add_node(pydot.Node(topic_id, label=node_label, shape='box', URL=node_url, target='_top'))
            graph.add_edge(pydot.Edge(self.__name, topic_id))

            for sub_name in topic.subscribers:
                subnode_id = 'sub_' + sub_name
                if sub_name not in node_names:
                    subnode_url = url_for('node.get_node_info', name=sub_name)
                    node_label = f"{sub_name}\\n{topic.subscribers[sub_name].uri}"
                    graph.add_node(pydot.Node(subnode_id, label=node_label, shape='oval', URL=subnode_url, target='_top'))
                    node_names.add(sub_name)
                graph.add_edge(pydot.Edge(topic_id, subnode_id))

        # Store and return graph
        self.__graph = graph
        return graph

    def svg(self):
        # Check if we can return an existing svg
        graph = self.graph()
        if (self.__svg is not None) and (self.__times['svg'] >= self.__times['graph']):
            return self.__svg

        # Remember time
        rospy.loginfo(f'Generating svg for node {self.__name}')
        self.__times['svg'] = datetime.now()

        # Store and return svg
        self.__svg = graph_to_svg(graph)
        return self.__svg


class Service:
    def __init__(self, name):
        self.__name = name
        try:
            self.__type = rosservice.get_service_type(name)
            self.__data_class = rosservice.get_service_class_by_name(name)
            self.__data_class_request = roslib.message.get_service_class(self.__type + 'Request')
        except Exception as e:
            rospy.logerr(str(e))
            self.__type = None
            self.__data_class = None
            self.__data_class_request = None
        self.__providers = dict()
        self.__graph = None
        self.__svg = None
        self.__times = {
            'state': datetime.now(),
            'graph': None,
            'svg': None
        }

    @property
    def name(self):
        return self.__name

    @property
    def type(self):
        return self.__type

    @property
    def providers(self):
        return self.__providers

    @property
    def data_class(self):
        return self.__data_class

    @property
    def request_class(self):
        return self.__data_class_request

    def call(self, args):
        try:
            return rospy.ServiceProxy(self.__name, self.__data_class)(args)
        except Exception as e:
            rospy.loginfo(e)
            raise e

    def clear(self):
        self.__providers.clear()
        self.__times['state'] = datetime.now()

    def addProvider(self, node):
        self.__providers[node.name] = node
        self.__times['state'] = datetime.now()

    def graph(self):
        # Check if we can return an existing graph
        if (self.__graph is not None) and (self.__times['graph'] >= self.__times['state']):
            return self.__graph

        # Remember time
        rospy.loginfo(f'Generating graph for service {self.__name}')
        self.__times['graph'] = datetime.now()

        # Generate new graph
        graph = pydot.Dot(graph_type='digraph', forcelabels=True, stylesheet='https://www.w3schools.com/w3css/4/w3.css')
        
        node_label = f"{self.__name}\\n{self.__type}"
        node_id = "srv_" + self.__name
        graph.add_node(pydot.Node(node_id, label=node_label, shape='box'))

        for _, node in self.__providers.items():
            node_url = url_for('node.get_node_info', name=node.name)
            graph.add_node(pydot.Node(node.name, shape='oval', URL=node_url, target='_top'))
            graph.add_edge(pydot.Edge(node.name, node_id))

        # Store and return graph
        self.__graph = graph
        return graph

    def svg(self):
        # Check if we can return an existing svg
        graph = self.graph()
        if (self.__svg is not None) and (self.__times['svg'] >= self.__times['graph']):
            return self.__svg

        # Remember time
        rospy.loginfo(f'Generating svg for node {self.__name}')
        self.__times['svg'] = datetime.now()

        # Store and return svg
        self.__svg = graph_to_svg(graph)
        return self.__svg


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ROSApi(metaclass=Singleton):
    def __init__(self):
        rospy.loginfo("Instance created")
        self.__system_state = None
        self.__nodes = dict()
        self.__topics = dict()
        self.__services = dict()
        self.__params = dict()
        self.__blacklisted_topics = set()
        self.__blacklisted_nodes = set()
        self.__graph = None
        self.__svg = None
        self.__master_pid = None
        self.__times = {
            'state': datetime.now(),
            'graph': None,
            'svg': None
        }

    @property
    def topics(self):
        return self.__topics
    
    @property
    def nodes(self):
        return self.__nodes

    @property
    def services(self):
        return self.__services

    @property
    def params(self):
        return self.__params
    
    def config(self, blacklisted_topics, blacklisted_nodes):
        self.__blacklisted_topics = set([name if name[0] == '/' else '/' + name for name in blacklisted_topics])
        self.__blacklisted_nodes = set([name if name[0] == '/' else '/' + name for name in blacklisted_nodes])

    def graph(self):
        # Check if we can return an existing graph
        if (self.__graph is not None) and (self.__times['graph'] >= self.__times['state']):
            return self.__graph

        # Remember time
        rospy.loginfo(f'Generating overview graph')
        self.__times['graph'] = datetime.now()

        # Generate new graph
        graph = pydot.Dot(graph_type='digraph', forcelabels=True)

        # Add one graph node for each ros node
        for node_name in ros.nodes:
            node_url = url_for('node.get_node_info', name=node_name)
            graph.add_node(pydot.Node(node_name, shape='oval', URL=node_url, target='_top'))
            
        # Iterate over topics, add a graph node for each topic and draw edges to subscribers and publishers
        for topic_name, topic in ros.topics.items():
            topic_id = 'topic_' + topic_name
            node_url = url_for('topic.get_topic_info', name=topic_name)
            node_label = f"{topic_name}\\n{topic.type}"
            graph.add_node(pydot.Node(topic_id, label=node_label, shape='box', URL=node_url, target='_top'))

            for node_name in topic.publishers:
                graph.add_edge(pydot.Edge(node_name, topic_id))

            for node_name in topic.subscribers:
                graph.add_edge(pydot.Edge(topic_id, node_name))

        # Store and return graph
        self.__graph = graph
        return graph

    def svg(self):
        # Check if we can return an existing svg
        graph = self.graph()
        if (self.__svg is not None) and (self.__times['svg'] >= self.__times['graph']):
            return self.__svg

        # Remember time
        rospy.loginfo(f'Generating overview svg')
        self.__times['svg'] = datetime.now()

        # Store and return svg
        self.__svg = graph_to_svg(graph)
        return self.__svg

    def get_topic(self, name):
        if name in self.__topics:
            return self.__topics[name]
        if '/'+name in self.__topics:
            return self.__topics['/'+name]
        return None

    def get_node(self, name):
        if name in self.__nodes:
            return self.__nodes[name]
        return None

    def get_service(self, name):
        if name in self.__services:
            return self.__services[name]
        return None

    def update(self):
        # get master
        master = rosgraph.Master('rosnode')

        # go through the master system state first
        try:
            pid = master.getPid()
            state = master.getSystemState()
        except socket.error:
            pid = None
            raise rosnode.ROSNodeIOException("Unable to communicate with master!")
        
        try:
            param_names = rosparam.list_params('')
        except rosparam.RosParamIOException:
            rospy.loginfo('Could not fetch parameter names from server')
        self.__params = sorted(set(['/'.join(param.split('/')[:2]) for param in param_names]))

        if state is None or state != self.__system_state or pid != self.__master_pid:
            # Clear state
            rospy.loginfo('Resetting nodes and services')
            self.__times['state'] = datetime.now()
            
            # Topics are not deleted so that we don't loose the message statistics
            self.__nodes.clear()
            self.__services.clear()
            if pid != self.__master_pid:
                rospy.loginfo("Resetting topics since roscore PID has changed")
                self.__master_pid = pid
                topic_names = []
                for name, topic in self.__topics.items():
                    topic_names.append(name)
                    topic.unregister()
                for name in topic_names:
                    del self.__topics[name]
            else:    
                for _, topic in self.__topics.items():
                    topic.clear()

            # Iterate over publisher topics and create nodes and topics if necessary
            for s in state[0]:
                topic_name = s[0]
                if topic_name not in self.__topics and topic_name not in self.__blacklisted_topics:
                    self.__topics[topic_name] = Topic(topic_name)
                for node_name in s[1]:
                    if node_name not in self.__blacklisted_nodes:
                        if node_name not in self.__nodes:
                            self.__nodes[node_name] = Node(node_name)
                        if topic_name not in self.__blacklisted_topics:
                            self.__topics[topic_name].addPublisher(self.__nodes[node_name])
                            self.__nodes[node_name].addPublication(self.__topics[topic_name])

            # Iterate over subscriber topics and create nodes and topics if necessary
            for s in state[1]:
                topic_name = s[0]
                if topic_name not in self.__topics and topic_name not in self.__blacklisted_topics:
                    self.__topics[topic_name] = Topic(topic_name)
                for node_name in s[1]:
                    if node_name not in self.__blacklisted_nodes:
                        if node_name not in self.__nodes:
                            self.__nodes[node_name] = Node(node_name)
                        if topic_name not in self.__blacklisted_topics:
                            self.__topics[topic_name].addSubscriber(self.__nodes[node_name])
                            self.__nodes[node_name].addSubscription(self.__topics[topic_name])

            # Delete blacklisted topics
            remove = [topic_name for topic_name in self.__topics if topic_name in self.__blacklisted_topics]
            for topic_name in remove: del self.__topics[topic_name]

            # Iterate over service topics and create nodes and services if necessary
            for s in state[2]:
                service_name = s[0]
                if service_name not in self.__services:
                    self.__services[service_name] = Service(service_name)
                self.__services[service_name].clear()
                for node_name in s[1]:
                    if node_name not in self.__blacklisted_nodes:
                        if node_name not in self.__nodes:
                            self.__nodes[node_name] = Node(node_name)
                        self.__services[service_name].addProvider(self.__nodes[node_name])
                        self.__nodes[node_name].addService(self.__services[service_name])
                # Remove service if it does not have a provider
                if len(self.__services[service_name].providers) == 0:
                    del self.__services[service_name]

            # Store system state
            self.__system_state = state

ros = ROSApi()
#rospy.init_node('ros_web_gui')
#_thread.start_new_thread(rospy.spin, ()) 
